import java.util.*;
import java.io.*;

/* This is an example actor that makes its decision on how to act
   based on a number of constants.  You can simply change the
   constants in this to try to create a better actor, or you can use
   it as a basis for writing your own actor, as many calls to the
   PlayerInterface functions are featured. It is currently not very
   intelligent, but it will make it to the end of the game */
public class ExampleActor extends PlayerActor {
    Vector cops, doughnuts, leaves, bulls;
    Position myPos;

    // these constants define two lower bounds for each resource - one
    // for the level at which you don't need that resource, and one
    // for the level at which you could use that resource.  Levels
    // below the latter are considered dire need.
    final int NONEED_ENERGY_LB   = 75;
    final int COULDUSE_ENERGY_LB = 50;
    final int NONEED_INSPIRATION_LB   = 75;
    final int COULDUSE_INSPIRATION_LB = 50;
    final int NONEED_DOUGHNUTS_LB   = 15;
    final int COULDUSE_DOUGHNUTS_LB =  2;

    // distance (manhattan) at which cops are close enough to consider
    // dropping a doughnut
    final int COPS_TOO_CLOSE = 7;

    // This is the maximum distance a doughnut can be away from a cop
    // and be considered a decoy (i.e. a doughnut that the cop could
    // be going after).  If a decoy exists when the player is near a
    // cop, no doughnut will be dropped.  In addition, all doughnuts
    // that are decoys are ignored by the player when scoring objects.
    final int MAX_DECOY_DISTANCE = 8;

    // The player is allowed to eat doughnuts for energy when he has
    // more than enough
    final int ENOUGH_DOUGHNUTS = 6;

    // The player will go directly to the clue when it is this close
    final int GO_TO_CLUE_DIST = 7;

    // The rest of the constants help determine the scores of close
    // bonus objects (NC = no clue present for player, C = clue
    // present for player)

    // These serve as base scores, so that the score is different
    // based on whether a clue is present (see scoring functions to
    // see how this is used).
    final double NC_BASE = 4.20;
    final double C_BASE = 4.20;

    // The following constants are factors to be used for scoring
    // purposes.  NC factors are used when no clue is present.  C
    // factors are used when a clue is prssent.  NONEED factors are
    // used when the resource level is equal to or above the NONEED
    // lower bound, defined above.  COULDUSE factors are used when
    // level is equal to or above the COULDUSE lower bound.  DIRENEED
    // factors are used when the level is below the COULDUSE lower
    // bound.

    // The distance factor is used as a gauge of how close something
    // should be before the player goes to get it.  The higher the
    // factor, the lower the score will be for the respective resource
    // (i.e. the closer it will have to be to score well).  This is
    // multiplicative (see scoring functions for how it is used).
    final double NC_BULL_NONEED_DISTANCE_FACTOR = 4.20;
    final double NC_BULL_COULDUSE_DISTANCE_FACTOR  =4.20;
    final double NC_BULL_DIRENEED_DISTANCE_FACTOR = 4.20;
    final double NC_JAY_NONEED_DISTANCE_FACTOR = 4.20;
    final double NC_JAY_COULDUSE_DISTANCE_FACTOR = 4.20;
    final double NC_JAY_DIRENEED_DISTANCE_FACTOR = 4.20;
    final double NC_DONUT_NONEED_DISTANCE_FACTOR = 4.20;
    final double NC_DONUT_COULDUSE_DISTANCE_FACTOR = 4.20;
    final double NC_DONUT_DIRENEED_DISTANCE_FACTOR = 4.20;

    final double C_BULL_NONEED_DISTANCE_FACTOR = 4.20;
    final double C_BULL_COULDUSE_DISTANCE_FACTOR = 4.20;
    final double C_BULL_DIRENEED_DISTANCE_FACTOR = 4.20;
    final double C_JAY_NONEED_DISTANCE_FACTOR = 4.20;
    final double C_JAY_COULDUSE_DISTANCE_FACTOR = 4.20;
    final double C_JAY_DIRENEED_DISTANCE_FACTOR = 4.20;
    final double C_DONUT_NONEED_DISTANCE_FACTOR = 4.20;
    final double C_DONUT_COULDUSE_DISTANCE_FACTOR = 4.20;
    final double C_DONUT_DIRENEED_DISTANCE_FACTOR = 4.20;

    // The need factor is used as a gauge of how much the lack of the
    // resource should outweigh how far away it is.  The higher the
    // need factor, the better the score will be for that resource
    // (i.e. it can be further away and still score well).  This is
    // additive (see scoring functions for how this is used).
    final double NC_BULL_NONEED_NEED_FACTOR = 4.20;
    final double NC_BULL_COULDUSE_NEED_FACTOR = 4.20;
    final double NC_BULL_DIRENEED_NEED_FACTOR = 4.20;
    final double NC_JAY_NONEED_NEED_FACTOR = 4.20;
    final double NC_JAY_COULDUSE_NEED_FACTOR = 4.20;
    final double NC_JAY_DIRENEED_NEED_FACTOR = 4.20;
    final double NC_DONUT_NONEED_NEED_FACTOR = 4.20;
    final double NC_DONUT_COULDUSE_NEED_FACTOR = 4.20;
    final double NC_DONUT_DIRENEED_NEED_FACTOR = 4.20;

    final double C_BULL_NONEED_NEED_FACTOR = 4.20;
    final double C_BULL_COULDUSE_NEED_FACTOR = 4.20;
    final double C_BULL_DIRENEED_NEED_FACTOR = 4.20;
    final double C_JAY_NONEED_NEED_FACTOR = 4.20;
    final double C_JAY_COULDUSE_NEED_FACTOR = 4.20;
    final double C_JAY_DIRENEED_NEED_FACTOR = 4.20;
    final double C_DONUT_NONEED_NEED_FACTOR = 4.20;
    final double C_DONUT_COULDUSE_NEED_FACTOR = 4.20;
    final double C_DONUT_DIRENEED_NEED_FACTOR = 4.20;
    
    public ExampleActor(PlayerInterface pi) {
	// this line should always be called
	super(pi);

	// initialize variables
	cops = new Vector();
	doughnuts = new Vector();
	bulls = new Vector();
	leaves = new Vector();
	myPos = pi.getMyPos();
    }

    // returns the manhattan distance from one grid position to
    // another (manhattan distance is defined as the number of squares
    // you must traverse to get there).  Walls are not taken into
    // account.
    public int mandist(int x1, int y1, int x2, int y2) {
	return Math.abs(x1 - x2) + Math.abs(y1 - y2);	
    }

    // returns true if nothing is blocking the player in the specified
    // direction
    public boolean isWayClear(int dir, PlayerInterface pi) {
	switch (dir) {
	case 0: return pi.isNorthClear();
	case 1: return pi.isEastClear();
	case 2: return pi.isSouthClear();
	case 3: return pi.isWestClear();
	default:
	    return false;
	}
    }

    // The next two functions return an action that moves the player
    // towards the specified position.  If multiple directions do
    // that, it chooses randomly among them.
    public int chooseRandomActionTowards(int x, int y, PlayerInterface pi) {
	return chooseRandomActionTowards(new Position(x,y), pi);
    }

    public int chooseRandomActionTowards(Position pos, PlayerInterface pi) {
	boolean [] dir = new boolean[4];
	int theDir = -1;
	for (int i=0;i<4;i++)
	    dir[i]=false;
	if (myPos.x < pos.x && isWayClear(1,pi) )
	    dir[1] = true;
	if (myPos.x > pos.x && isWayClear(3,pi))
	    dir[3] = true;
	if (myPos.y < pos.y && isWayClear(2,pi))
	    dir[2] = true;
	if (myPos.y > pos.y && isWayClear(0,pi))
	    dir[0] = true;
	int roll = rg.nextInt(4);
	for (int i=0;i<4;i++) {
	    if (dir[(roll+i)%4]) {
		theDir = (roll+i)%4;
		break;
	    }
	}
	if (theDir > -1) 
	    return theDir;
	else {
	    return chooseRandomAction();
	}
    }

    // returns a random action that is not blocked
    public int chooseRandomClearAction(PlayerInterface pi) {
	boolean [] dir = new boolean[4];
	for (int i=0;i<4;i++)
	    dir[i]=false;
	int theDir = -1;
	if (isWayClear(1,pi) )
	    dir[1] = true;
	if (isWayClear(3,pi))
	    dir[3] = true;
	if (isWayClear(2,pi))
	    dir[2] = true;
	if (isWayClear(0,pi))
	    dir[0] = true;
	int roll = rg.nextInt(4); 
	for (int i=0;i<4;i++) {
	    if (dir[(roll+i)%4]) {
		theDir = (roll+i)%4;
		break;
	    }
	}
	if (theDir > -1) 
	    return theDir;
	else 
	    return chooseRandomAction();
    }

    // returns a completely random action
    public int chooseRandomAction() {
	float roll = rg.nextFloat();
	if (roll < 0.25)
	    return 0;
	else if (roll < 0.50)
	    return 1;
	else if (roll < 0.75)
	    return 2;
	else
	    return 3;
    }

    // returns the distance penalty incurred for getting this object
    // before getting the clue (i.e. the number of additional steps
    // required)
    public int penaltyToClue(PerceptObject po) {
	Position cluePos = pi.getCluePos();
	if (cluePos == null)
	    return 0;
	else 
	    return ((mandist(myPos.x, myPos.y, po.x, po.y) + 
		     pi.distToClue(po.x,po.y))
		    - pi.distToClue(myPos.x, myPos.y));
    }

    // The next six functions are the scoring functions for the three
    // resources, both for when a clue is present and when no clue is
    // present (i.e player is thinking, or solving last clue).  The
    // objects are scored in the following manner: 
    //(Base{Clue/NoClue} + (NeedForThisCase -
    //(distance/penalty)*distancefactorForThisCase)

    public double scoreNCJay(PerceptObject bestJ) {
	double need, distfact;
	int dist = mandist(myPos.x, myPos.y, bestJ.x, bestJ.y);
	if (pi.getInspiration() >= NONEED_INSPIRATION_LB) {
	    need = NC_JAY_NONEED_NEED_FACTOR;
	    distfact = NC_JAY_NONEED_DISTANCE_FACTOR;
	} else if (pi.getInspiration() >= COULDUSE_INSPIRATION_LB) {
	    need = NC_JAY_COULDUSE_NEED_FACTOR;
	    distfact = NC_JAY_COULDUSE_DISTANCE_FACTOR;
	} else {
	    need = NC_JAY_DIRENEED_NEED_FACTOR;
	    distfact = NC_JAY_DIRENEED_DISTANCE_FACTOR;
	}
	return NC_BASE + need - (dist*distfact);
    }

    public double scoreNCBull(PerceptObject bestB) {
	double need, distfact;
	int dist = mandist(myPos.x, myPos.y, bestB.x, bestB.y);
	if (pi.getEnergy() >= NONEED_ENERGY_LB) {
	    need = NC_BULL_NONEED_NEED_FACTOR;
	    distfact = NC_BULL_NONEED_DISTANCE_FACTOR;
	} else if (pi.getEnergy() >= COULDUSE_ENERGY_LB) {
	    need = NC_BULL_COULDUSE_NEED_FACTOR;
	    distfact = NC_BULL_COULDUSE_DISTANCE_FACTOR;
	} else {
	    need = NC_BULL_DIRENEED_NEED_FACTOR;
	    distfact = NC_BULL_DIRENEED_DISTANCE_FACTOR;
	}
	return NC_BASE + need - (dist*distfact);
    }

    public double scoreNCDonut(PerceptObject bestD) {
	double need, distfact;
	int dist = mandist(myPos.x, myPos.y, bestD.x, bestD.y);
	if (pi.getDoughnutCount() >= NONEED_DOUGHNUTS_LB) {
	    need = NC_DONUT_NONEED_NEED_FACTOR;
	    distfact = NC_DONUT_NONEED_DISTANCE_FACTOR;
	} else if (pi.getDoughnutCount() >= COULDUSE_DOUGHNUTS_LB) {
	    need = NC_DONUT_COULDUSE_NEED_FACTOR;
	    distfact = NC_DONUT_COULDUSE_DISTANCE_FACTOR;
	} else {
	    need = NC_DONUT_DIRENEED_NEED_FACTOR;
	    distfact = NC_DONUT_DIRENEED_DISTANCE_FACTOR;
	}
	return NC_BASE + need - (dist*distfact);
    }

    public double scoreCJay(PerceptObject bestJ) {
	double need, distfact;
	int dist = penaltyToClue(bestJ);
	if (pi.getInspiration() >= NONEED_INSPIRATION_LB) {
	    need = C_JAY_NONEED_NEED_FACTOR;
	    distfact = C_JAY_NONEED_DISTANCE_FACTOR;
	} else if (pi.getInspiration() >= COULDUSE_INSPIRATION_LB) {
	    need = C_JAY_COULDUSE_NEED_FACTOR;
	    distfact = C_JAY_COULDUSE_DISTANCE_FACTOR;
	} else {
	    need = C_JAY_DIRENEED_NEED_FACTOR;
	    distfact = C_JAY_DIRENEED_DISTANCE_FACTOR;
	}
	return C_BASE + need - (dist*distfact);
    }

    public double scoreCBull(PerceptObject bestB) {
	double need, distfact;
	int dist = penaltyToClue(bestB);
	if (pi.getEnergy() >= NONEED_ENERGY_LB) {
	    need = C_BULL_NONEED_NEED_FACTOR;
	    distfact = C_BULL_NONEED_DISTANCE_FACTOR;
	} else if (pi.getEnergy() >= COULDUSE_ENERGY_LB) {
	    need = C_BULL_COULDUSE_NEED_FACTOR;
	    distfact = C_BULL_COULDUSE_DISTANCE_FACTOR;
	} else {
	    need = C_BULL_DIRENEED_NEED_FACTOR;
	    distfact = C_BULL_DIRENEED_DISTANCE_FACTOR;
	}
	return C_BASE + need - (dist*distfact);
    }

    public double scoreCDonut(PerceptObject bestD) {
	double need, distfact;
	int dist = penaltyToClue(bestD);
	if (pi.getDoughnutCount() >= NONEED_DOUGHNUTS_LB) {
	    need = C_DONUT_NONEED_NEED_FACTOR;
	    distfact = C_DONUT_NONEED_DISTANCE_FACTOR;
	} else if (pi.getDoughnutCount() >= COULDUSE_DOUGHNUTS_LB) {
	    need = C_DONUT_COULDUSE_NEED_FACTOR;
	    distfact = C_DONUT_COULDUSE_DISTANCE_FACTOR;
	} else {
	    need = C_DONUT_DIRENEED_NEED_FACTOR;
	    distfact = C_DONUT_DIRENEED_DISTANCE_FACTOR;
	}
	return C_BASE + need - (dist*distfact);
    }

    // returns the object that causes you to incur the smallest
    // distance penalty if you get it before getting the clue
    public PerceptObject getBestToClue(Vector v) {
	int bestPenalty = 9999;
	PerceptObject bestPO = null;
	for (ListIterator i = v.listIterator(0); i.hasNext();) {
	    PerceptObject po = (PerceptObject)(i.next());
	    // ignore doughnuts close to cops
	    if (po.type == PerceptObject.DOUGHNUT && isADecoy(po))
		continue;
	    int lp = penaltyToClue(po);
	    if (lp < bestPenalty) {
		bestPenalty = lp;
		bestPO = po;
	    }
	}
	return bestPO;
    }

    public int ClueAction() {
	// if clue is close enough, just go get it
	if (pi.distToClue(myPos.x, myPos.y) <= GO_TO_CLUE_DIST) {
	    return pi.actToClue(myPos.x, myPos.y);
	} else {
	    PerceptObject bestDN, bestJ, bestRB;
	    double dnScore, jScore, rbScore;
	    // find the bonus objects that are most on the
	    // way to clue
	    bestDN = getBestToClue(doughnuts);
	    bestJ = getBestToClue(leaves);
	    bestRB = getBestToClue(bulls);
	    dnScore = jScore = rbScore = -9999;
	    // score those bonus objects, using C
	    // constants as a guide
	    if (bestDN != null)
		dnScore = scoreCDonut(bestDN);
	    if (bestJ != null)
		jScore  = scoreCJay(bestJ);
	    if (bestRB != null)
		rbScore = scoreCBull(bestRB);
	    
	    // move towards the one with the best score,
	    // assuming it has a positive score.
	    // otherwise, just move towards the clue
	    if (jScore > 0 && jScore >= dnScore && jScore >= rbScore) {
		return chooseRandomActionTowards(bestJ.x, bestJ.y, pi);
	    } else if (dnScore > 0 && dnScore >= jScore && 
		       dnScore >= rbScore) {
		return chooseRandomActionTowards(bestDN.x, bestDN.y, pi);
	    } else if (rbScore > 0 && rbScore >= jScore && 
		       rbScore >= dnScore) {
		return chooseRandomActionTowards(bestRB.x, bestRB.y, pi);
	    } else {
		return pi.actToClue(myPos.x, myPos.y);
	    }
	}
    }

    // returns true if the doughnut is a decoy (a cop is close by)
    public boolean isADecoy(PerceptObject doughnut) {
	for (ListIterator i = cops.listIterator(); i.hasNext();) {
		PerceptObject c = (PerceptObject)(i.next());
		if (mandist(doughnut.x, doughnut.y, c.x, c.y) <= 
		    MAX_DECOY_DISTANCE)
		    return true;
	}
	return false;
    }

    public int NoClueAction() {
	// Get the best of each bonus object (usually the closest)
	PerceptObject bestB = (PerceptObject) bulls.get(0);
	PerceptObject bestJ =  (PerceptObject) leaves.get(0);
	// for doughnuts, take the closest that isn't in danger of
	// being eaten by a cop
	PerceptObject bestD = null;
	for (ListIterator i = doughnuts.listIterator(); i.hasNext();) {
	    bestD = (PerceptObject)(i.next());
	    if (!isADecoy(bestD))
		break;
	}

	double bestJScore, bestBScore, bestDScore;
	bestJScore = bestBScore = bestDScore = -9999;

	// Now score those bonus objects, using the NC
	// constants as a guide
	if (bestJ != null) 
	    bestJScore = scoreNCJay(bestJ);
	if (bestB != null)
	    bestBScore = scoreNCBull(bestB);
	if (bestD != null)
	    bestDScore = scoreNCDonut(bestD);
	
	// move towards the one with the best score,
	// assuming it has a positive score.
	// otherwise, just move randomly
	if (bestJ != null &&
	    bestJScore >= bestBScore && 
	    bestJScore >= bestDScore && 
	    bestJScore > 0) {
	    return chooseRandomActionTowards(bestJ.x, bestJ.y, pi);
	} else if (bestB != null &&
		   bestBScore >= bestJScore && 
		   bestBScore >= bestDScore && 
		   bestBScore > 0) {
	    return chooseRandomActionTowards(bestB.x, bestB.y, pi);
	} else if (bestD != null &&	
		   bestDScore >= bestBScore && 
		   bestDScore >= bestJScore &&
		   bestDScore > 0) {
	    return chooseRandomActionTowards(bestD.x, bestD.y, pi);
	} else {
	    return chooseRandomAction();
	}
    }

    public void checkForDrop() {
	//go through cops and see if there's one too close that's not
	//eating
	for (ListIterator i = cops.listIterator(0); i.hasNext();) {
	    PerceptObject po = (PerceptObject)(i.next());
	    if (mandist(myPos.x, myPos.y, po.x, po.y) <= COPS_TOO_CLOSE) {

		// see if there's a doughnut the cop could be going
		// after already on the board (decoy)
		boolean decoy = false;
		
		// if the cop is eating, there is a virtual decoy
		if (po.special > -1)
		    decoy = true;
		else {
		    // iterate through doughnuts
		    for (ListIterator j = doughnuts.listIterator(0); 
			 j.hasNext();) {
			PerceptObject dn = (PerceptObject)(j.next());
			
			// if a doughnut is close enough to cop, we don't
			// need to drop one
			if (mandist(dn.x, dn.y, po.x, po.y) <= 
			    MAX_DECOY_DISTANCE) {
			    decoy = true;
			    break;
			}
		    }
		}
		if (!decoy) {
		    pi.drop();
		    break;
		}
	    }
	    else
		break;
	}
    }

    // This agent figures out what to do every turn based on the
    // way constants are set.  It scores close bonus objects based
    // on those constants and decide when to move towards them.
    // It also decides when to eat and drop doughnuts based on
    // constants.
    public void act() {
	// get my position
	myPos = pi.getMyPos();

	// get other objects on board, except other players
	pi.getCops(cops);
	pi.getDoughnuts(doughnuts);
	pi.getLeaves(leaves);
	pi.getBulls(bulls);

	// if we could use energy and we have enough doughnuts, eat
	if (pi.getEnergy() < NONEED_ENERGY_LB && 
	    pi.getDoughnutCount() > ENOUGH_DOUGHNUTS) {
	    pi.eat();
	}

	// check whether we want to drop a doughnut first
	if (pi.getDoughnutCount() > 0)
	    checkForDrop();

	// decide what move to make, based on the board configuration,
	// the resource levels, the constants, and whether there is a
	// clue to go to

	// if the player is held, don't try to move
	if (pi.getHeldTime() > 0) {
	    pi.setDir(LocalMap.STOP);
	}
	// with low probability, make a totally random move.
	// Hopefully, this keeps us from ever getting stuck
	else if (rg.nextFloat() < 0.01) {
	    pi.setDir(chooseRandomAction());
	}
	else {              	    // Now, decide what to do.
	    int planDir = -1;
	    Position cluePos = pi.getCluePos();
	    if (cluePos == null) {	    // if there's no clue
		planDir = NoClueAction();
	    }
	    else {                        // there is a clue to go get!
		planDir = ClueAction();
	    }
	    if (planDir > -1) {
		// make sure the path is clear.  If not, move
		// randomly in a direction that is clear.
		if (isWayClear(planDir, pi))
		    pi.setDir(planDir);
		else {
		    pi.setDir(chooseRandomClearAction(pi));
		}
	    } else { // by default, move randomly 
		pi.setDir(chooseRandomAction());
	    }
	}
    }
    
}

