import java.util.*;

/** This is the class that the PlayerActor can use to get at the 
    features of the world. */

public class PlayerInterface {
    private Player player;
    private PlayerPercept playerPercept;
    private Position tPos;

    /** Not for you to use.  Used by the simulator and passed to PlayerActor */
    public PlayerInterface(Player p, PlayerPercept pp) {
	player = p; playerPercept = pp;
	tPos = new Position();
    }

    /** getMyPos() <BR>
	Returns your agent's current location as a "Position" object.  
	This is a simple object comprised of an x and y member variable, which you
	can access.  x,y represents the cell where the agent currently is, with x
	being horizontal distance from the upper left, and y being the vertical
	distance 
    */
    public  Position getMyPos() {
	return playerPercept.myPos;
    }
    /** getMyPreviousPos() <BR>
	Like the above, but returns the agent's position on the previous time step. */
    public  Position getMyPreviousPos() {
	return playerPercept.myPreviousPos;
    }
    /** getCluePos() <BR>
	Returns the position of your next clue, if it exists.  If you have not yet figured out the 
	location of the clue, it returns null.
    */
    public Position getCluePos() {
	return playerPercept.cluePos;
    } 

    /** getTimeElapsed() <BR>
	Returns the amount of game time elapsed in minutes.
    */
    public int getTimeElapsed() {
	return playerPercept.time;
    }

    /** getNumCluesSolved() <BR>
	Returns the number of clues that this player has solved.
    */
    public int getNumCluesSolved() {
	return playerPercept.nclues;
    }

    /** getDoughnutCount() <BR>
	Returns the number of doughnuts this player has.
    */
    public int getDoughnutCount() {
	return player.doughnutCount;
    }

    /** getEnergy() <BR>
	returns the amount of energy this player has.
    */
    public int getEnergy() {
	return player.energy;
    }

    /** getInspiration() <BR>
	returns the amount of inspiration this player has.
    */
    public int getInspiration() {
	return player.inspiration;
    }

    /** getThinkingTime() <BR>
	returns the amount of time that the player has been thinking
	trying to solve the current clue.  It returns 0 if the player
	has already solved the clue (but may return 0 if the player
	has just gotten the clue as well, as the player hasn't thought
	about the clue yet).  It's best to use getCluePos() first, and
	if it's null, then thinking time makes sense, otherwise it's 
	not really meaningful.
    */
    public int getThinkingTime() {
	return player.thinking;
    }

    /** getHeldTime() <BR>
	returns the amount of time that the player has been held by a cop.
	If this is 0, then the player is not currently held.
    */
    public int getHeldTime() {
	return player.held();
    }

    /** getObjectAt(int x, int y)  <BR>
	This returns the PerceptObject for the object at 
	x,y in the world.  If there is nothing there, it returns null.  If the 
	thing there is a clue, it also returns null.  
    */
    public PerceptObject getObjectAt(int x, int y) {
	MazeObject mo = playerPercept.world[x][y];
	if (mo == null || mo instanceof Clue || (x == player.x() && y==player.y()))
	    return null;
	else
	    return mo.po;
    }

    /** getDoughnuts(Vector v) <BR> This function takes in a vector v, 
	clears it, and fills it with PerceptObjects for the doughnuts that can be
	observed in the maze.  The Percept Object is a simple class
	that holds the position and previous position of objects that
	the player can see.   The objects are ordered in order of closeness to
	the agent.  Note that the special field in the PerceptObjects that are put into v is used to show the id of the team which dropped the doughnut (-1 = not dropped by team)
    */
    public void getDoughnuts(Vector v) {
	v.clear();
	for (ListIterator i = playerPercept.doughnuts.listIterator(0); i.hasNext();) {
	    ObjectTracker ot = (ObjectTracker)(i.next());
	    v.add(ot.po);
	}
    }

    /** getCops(Vector v) <BR> 
	Like getDoughnuts, but for cops.  Note that the special field in the 
	PerceptObjects that are put into v is used to show how long this cop has been eating (-1 = not eating).
    */
    public void getCops(Vector v) {
	v.clear();
	for (ListIterator i = playerPercept.cops.listIterator(0); i.hasNext();) {
	    ObjectTracker ot = (ObjectTracker)(i.next());
	    v.add(ot.po);
	}
    }

    /** getLeaves(Vector v) <BR> 
	Like getDoughnuts, but for jays. 
    */
    public void getLeaves(Vector v) {
	v.clear();
	for (ListIterator i = playerPercept.leaves.listIterator(0); i.hasNext();) {
	    ObjectTracker ot = (ObjectTracker)(i.next());
	    v.add(ot.po);
	}
    }

    /** getBulls(Vector v) <BR> 
	Like getDoughnuts, but for red bulls.
    */
    public void getBulls(Vector v) {
	v.clear();
	for (ListIterator i = playerPercept.bulls.listIterator(0); i.hasNext();) {
	    ObjectTracker ot = (ObjectTracker)(i.next());
	    v.add(ot.po);
	}
    }

    /** getOtherPlayers(Vector v) <BR> 
	Like getDoughnuts, but for other players. 
	Note that the special field in the 
	PerceptObjects that are put into v is used to show long the given player has been held (-1 = not held).
    */
    public void getOtherPlayers(Vector v) {
	v.clear();
	for (ListIterator i = playerPercept.others.listIterator(0); i.hasNext();) {
	    ObjectTracker ot = (ObjectTracker)(i.next());
	    v.add(ot.po);
	}
    }


    /** isNorthClear() <BR>
	Simply returns a boolean as to whether there is a wall or other player to the north
    */
    public boolean isNorthClear() {
	return playerPercept.clearUp;
    }
    /** isEastClear() <BR>
	Simply returns a boolean as to whether there is a wall or other player to the east.
    */
    public boolean isEastClear() {
	return playerPercept.clearRight;
    }
    /** isSouthClear() <BR>
	Simply returns a boolean as to whether there is a wall or other player to the south.
    */
    public boolean isSouthClear() {
	return playerPercept.clearDown;
    }
    /** isWestClear() <BR>
	Simply returns a boolean as to whether there is a wall or other player to the west.
    */
    public boolean isWestClear() {
	return playerPercept.clearLeft;
    }

    /** distToClue(x,y) <BR>
	returns the actual distance to the clue, taking walls (but not any moving obstacles) into 
	account.  This function returns -1 if there is no clue in the maze for this player...
    */
    public int distToClue(int x, int y) {
	tPos.x = x; tPos.y=y;
	if (playerPercept.cluePos != null)
	    return playerPercept.cluePlan.getDist(tPos);
	else
	    return -1;
    }

    /** actToClue(x,y) <BR>
	This returns the action that will bring the agent closer to the clue. 
    */
    public int actToClue(int x, int y) {
	tPos.x = x; tPos.y=y;
	if (playerPercept.cluePos != null)
	    return playerPercept.cluePlan.getAction(tPos);
	else
	    return LocalMap.STOP;
    }

    /** setDir(vdir) <BR>
	This function sets the player's direction to dir.  
	It uses the standard convention of 0=North, 1=east, 2=south, 3=west, 4=stop.
    */
    public void setDir(int dir) {
	player.command = dir;
	if (dir == LocalMap.STOP) {
	    player.command = -1;
	}
    }

    /** eat() <BR>
	Tell the player to eat a doughnut.
    */
    public void eat() {
	player.eatDoughnut = true;
    }

    /** drop() <BR>
	Tell the player to drop a doughnut.
    */
    public void drop() {
	player.dropDoughnut = true;
    }

}
